@@ -6,8 +6,8 @@
 
       Jane Street Holding, LLC
       Author: Markus Mottl
-      email: mmottl\@janestcapital.com
-      WWW: http://www.janestcapital.com/ocaml
+      email: mmottl\@janestreet.com
+      WWW: http://www.janestreet.com/ocaml
 
    This file is derived from file "pa_tywith.ml" of version 0.45 of the
    library "Tywith".
@@ -85,7 +85,6 @@ let set_conv_path conv_path =
     conv_path_ref := Path (conv_path, [conv_path])
   else failwith "Pa_type_conv: module name set twice"
 
-
 let () = if !Sys.interactive then set_conv_path "Toplevel"
 
 let push_conv_path mod_name =
@@ -320,10 +319,34 @@ let gen_derived_exn_sigs _loc tp drvs =
 
 open Syntax
 
+let is_prefix ~prefix x =
+  let prefix_len = String.length prefix in
+  String.length x >= prefix_len && prefix = String.sub x 0 prefix_len
+
+let chop_prefix ~prefix x =
+  if is_prefix ~prefix x then
+    let prefix_len = String.length prefix in
+    Some (String.sub x prefix_len (String.length x - prefix_len))
+  else None
+
+let get_default_path _loc =
+  try
+    let prefix = Sys.getenv "TYPE_CONV_ROOT" in
+    match chop_prefix ~prefix (Loc.file_name (Loc.make_absolute _loc)) with
+    | Some x -> x ^ "#"
+    | None -> Loc.file_name _loc
+  with _ -> Loc.file_name _loc
+
+let set_conv_path_if_not_set _loc =
+  if !conv_path_ref = Not_initialized || !Sys.interactive then
+    let conv_path = get_default_path _loc in
+    conv_path_ref := Path (conv_path, [conv_path])
+
 let found_module_name =
   Gram.Entry.of_parser "found_module_name" (fun strm ->
     match Stream.npeek 1 strm with
-    | [(UIDENT name, _)] ->
+    | [(UIDENT name, _loc)] ->
+        set_conv_path_if_not_set _loc;
         push_conv_path name;
         Stream.junk strm;
         name
@@ -339,13 +362,12 @@ let rec fetch_generator_arg paren_count strm =
   | x -> x :: fetch_generator_arg paren_count strm
 
 let generator_arg =
-  Gram.Entry.of_parser "generator_arg"
-    (fun strm ->
-       match Stream.peek strm with
-       | Some(KEYWORD "(", _) ->
-           Stream.junk strm;
-           Some (fetch_generator_arg 1 strm)
-       | _ -> None)
+  Gram.Entry.of_parser "generator_arg" (fun strm ->
+    match Stream.peek strm with
+    | Some (KEYWORD "(", _) ->
+        Stream.junk strm;
+        Some (fetch_generator_arg 1 strm)
+    | _ -> None)
 
 DELETE_RULE Gram str_item: "module"; a_UIDENT; module_binding0 END;
 
@@ -353,47 +375,49 @@ EXTEND Gram
   GLOBAL: str_item sig_item;
 
   str_item:
-  [[
-    "TYPE_CONV_PATH"; conv_path = STRING ->
-      set_conv_path conv_path;
-      <:str_item< >>
-  ]];
+    [[
+      "TYPE_CONV_PATH"; conv_path = STRING ->
+        set_conv_path conv_path;
+        <:str_item< >>
+    ]];
 
   generator: [[ id = LIDENT; arg = generator_arg -> (id, arg) ]];
 
   str_item:
-   [[
-     "type"; tds = type_declaration; "with";
-     drvs = LIST1 generator SEP "," ->
-       <:str_item< type $tds$; $gen_derived_defs _loc tds drvs$ >>
-  ]];
+    [[
+      "type"; tds = type_declaration; "with"; drvs = LIST1 generator SEP "," ->
+        set_conv_path_if_not_set _loc;
+        <:str_item< type $tds$; $gen_derived_defs _loc tds drvs$ >>
+    ]];
 
   str_item:
-   [[
-     "exception"; tds = constructor_declaration; "with";
-     drvs = LIST1 generator SEP "," ->
-       <:str_item< exception $tds$; $gen_derived_exn_defs _loc tds drvs$ >>
-  ]];
+    [[
+      "exception"; tds = constructor_declaration; "with";
+      drvs = LIST1 generator SEP "," ->
+        set_conv_path_if_not_set _loc;
+        <:str_item< exception $tds$; $gen_derived_exn_defs _loc tds drvs$ >>
+    ]];
 
   sig_item:
-   [[
-     "type"; tds = type_declaration; "with";
-     drvs = LIST1 generator SEP "," ->
-       <:sig_item< type $tds$; $gen_derived_sigs _loc tds drvs$ >>
-  ]];
+    [[
+      "type"; tds = type_declaration; "with"; drvs = LIST1 generator SEP "," ->
+        set_conv_path_if_not_set _loc;
+        <:sig_item< type $tds$; $gen_derived_sigs _loc tds drvs$ >>
+    ]];
 
   sig_item:
    [[
      "exception"; cd = constructor_declaration; "with";
      drvs = LIST1 generator SEP "," ->
+       set_conv_path_if_not_set _loc;
        <:sig_item< exception $cd$; $gen_derived_exn_sigs _loc cd drvs$ >>
-  ]];
+    ]];
 
   str_item:
-  [[
-    "module"; i = found_module_name; mb = module_binding0 ->
-      pop_conv_path ();
-      <:str_item< module $i$ = $mb$ >>
-  ]];
+    [[
+      "module"; i = found_module_name; mb = module_binding0 ->
+        pop_conv_path ();
+        <:str_item< module $i$ = $mb$ >>
+    ]];
 
 END
